"""Files over LAN.
Written by Matthew Timms for MonashUAS

Send/Receives files, or polls directory until user kills process.

Usage:
  folan.py listen <ip:port> [--save_path DIR_PATH] [options]
  folan.py send files <ip:port> [options] <file_path>...
  folan.py send dir <ip:port> <dir_path> [options]
  folan.py (-h | --help)

Example:
  folan.py listen 196.168.0.13:40000 -s . --limit=10
  folan.py send dir 10.100.192.15:5555 imgs/ --stayalive --limit=10

Options:
  -h --help                 Shows this screen.
  --stayalive               Continues to poll directory and send new files
  --limit LEN_FILES         Limits number of files to receive before closing
  -s, --save_path DIR_PATH  Path to saving directory [default: folan_dest/].
"""

from __future__ import print_function
import socket
import struct
import os
import time
import threading

__all__ = ['folan']
__version__ = '1.1.2'


class Client(object):
    def __init__(self, ip, port, timeout=3, debug=False):
        self.dest = (ip, port)
        self.timeout = timeout
        self.debug = debug
        self.sock = None
        self.buffer_size = 4096
        self.files_sent_len = 0
        self.printer = None

    def connect(self):
        """ Returns 1 if successfully connects with target host else 0 after timeout """
        if self.printer is not None and self.printer.finish_time is None:
            self.printer.release_violently()
        self._print_dbg("\r# Trying to Connect... ", end='')
        self.sock = socket.socket()
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)  # Allow address reuse
        self.sock.settimeout(self.timeout)
        try:
            self.sock.connect(self.dest)
        except socket.error as error:
            self._print_dbg("Could not connect: {}".format(error), end=' ')
            return 0
        self._print_dbg("Connection made")
        return 1

    def close(self):
        self.sock.close()

    def send_file(self, filepath):
        self.printer = PrintyMcPrintington.current_file(filepath)
        _, filename = os.path.split(filepath)
        filename_size = struct.pack('I', len(filename))
        self.sock.send(filename_size)  # Length of next expected pkt sent
        self.sock.send(filename.encode())

        filesize = os.path.getsize(filepath)
        sizepack = struct.pack('I', filesize)
        self.sock.send(sizepack)

        with open(filepath, 'rb') as f:
            buff = f.read(self.buffer_size)
            while buff:
                self.sock.send(buff)
                buff = f.read(self.buffer_size)
                self.printer.pkts_moved += 1

        self.sock.recv(13).decode()  # TODO
        self.printer.release_gracefully()
        self.files_sent_len += 1

    def _print_dbg(self, *args, **kwargs):
        if self.debug:
            print(*args, **kwargs)


class Server(object):
    def __init__(self, ip, port, timeout=3, debug=False):
        self.dest = (ip, port)
        self.timeout = timeout
        self.debug = debug
        self.sock = None
        self.conn = None
        self.buffer_size = 4096
        self.files_recv_len = 0
        self.printer = None

    def connect(self):
        """ Returns 1 if successfully connects with target host else 0 after timeout """
        if self.printer is not None and self.printer.finish_time is None:
            self.printer.release_violently()
        self._print_dbg("\r# Trying to Connect... ", end='')
        try:
            self.sock = socket.socket()
            self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self.sock.bind(self.dest)
            self.sock.listen(self.timeout)
        except socket.error as message:
            self._print_dbg("Could not open socket: {}".format(message), end=' ')
            return 0
        self.sock.settimeout(self.timeout)
        try:
            self.conn, addr = self.sock.accept()
        except socket.timeout:
            return 0
        self._print_dbg("Got connection from {}".format(addr))
        return 1

    def close(self):
        self.conn.close()
        self.sock.close()

    def recv_file(self, save_directory=''):
        """ Receives file and writes into save_directory """
        filename_size = self.conn.recv(struct.calcsize("!I"))
        if len(filename_size) == 0:
            raise socket.error("Socket closure")
        filename_size = struct.unpack('I', filename_size)[0]
        filename = self.conn.recv(filename_size).decode()

        sizepack = self.conn.recv(struct.calcsize('!I'))
        filesize = struct.unpack('I', sizepack)[0]
        self.printer = PrintyMcPrintington(filename, filesize)

        filepath = ''.join([save_directory, filename])
        with open(filepath, 'wb') as f:
            if filesize:
                while True:
                    data = self.conn.recv(self.buffer_size)
                    if len(data) == 0:
                        raise socket.error("Socket closure")
                    self.printer.pkts_moved += 1
                    f.write(data)
                    if f.tell() == filesize:
                        break
        self.conn.send("File received".encode())
        self.files_recv_len += 1
        self.printer.release_gracefully()

    def _print_dbg(self, *args, **kwargs):
        if self.debug:
            print(*args, **kwargs)


class PrintyMcPrintington(object):
    def __init__(self, filename, filesize, allow_printing=True, progress_bar_len=30, socket_buffer_size=4096):
        self.filename = filename
        self.filesize = filesize / 1024  # Bytes2Kilobytes
        self.progress_bar_len = progress_bar_len
        self.socket_buffer_size = socket_buffer_size  # Bytes
        self.allow_printing = allow_printing
        self.progress_ref = filesize / self.progress_bar_len / socket_buffer_size  # ratio of pkts moved to '=' icons
        self.start_time = time.time()
        self.finish_time = None
        self.stayalive = True
        self.pkts_moved = 0
        self._print_thread = threading.Thread(target=self.print_thread).start()

    def print_thread(self):
        while self.stayalive:
            if self.progress_ref == 0:
                progress = 0
            else:
                progress = int(self.pkts_moved / self.progress_ref)
            output = '\r%s\t|%s%s| %i/%iKB  elapse:%is' % (
                self.filename, '=' * progress, ' ' * (self.progress_bar_len - progress),
                self.pkts_moved * self.socket_buffer_size / 1024, self.filesize, time.time() - self.start_time)
            print(output, end='')
            time.sleep(0.1)

    @classmethod
    def current_file(cls, filepath):
        _, filename = os.path.split(filepath)
        filesize = os.path.getsize(filepath)
        return PrintyMcPrintington(filename, filesize)

    def release_gracefully(self):
        self.stayalive = False
        self.finish_time = time.time()
        elapse_time = self.finish_time - self.start_time
        while self._print_thread is not None:
            pass
        output = '\r%s\t|%s| size:%iKB  elapse:%is  rate:%.1f KB/s' % (
            self.filename, '=' * self.progress_bar_len, self.filesize, elapse_time, self.filesize / elapse_time)
        print(output)

    def release_violently(self):
        self.stayalive = False
        self.finish_time = time.time()
        while self._print_thread is not None:
            pass
        progress = int(self.pkts_moved / self.progress_ref)
        output = '\r%s\t|%sX%s| Connection broken!' % (
            self.filename, '=' * progress, ' ' * (self.progress_bar_len - progress - 1),)
        print(output)

    def print_decorator(self, some_function):
        def wrapper(*args, **kwargs):
            if self.allow_printing:
                some_function(*args, **kwargs)
        return wrapper


def main(args):
    if not isinstance(args, dict):
        raise TypeError("args expected as dictionary")
    
    ip, port = args['<ip:port>'].split(':')
    if ip == '\'\'':  # docopt's response to '' input
        ip = ''  # default interface
    port = int(port)

    if args['listen']:
        server = Server(ip, port, debug=True)

        save_directory = args['--save_path']
        if save_directory[-1] != '/':
            save_directory += '/'
        if args['--limit'] is not None:
            file_limit = int(args['--limit'])
        else:
            file_limit = 0

        if not os.path.exists(save_directory):
            os.makedirs(save_directory)

        while not server.connect():
            pass

        while True:
            if server.files_recv_len == file_limit and file_limit > 0:
                break

            try:
                server.recv_file(save_directory)
            except KeyboardInterrupt:
                print('KeyboardInterrupt')
                break
            except socket.error:
                while not server.connect():  # reconnect
                    pass

    elif args['send']:
        client = Client(ip, port, debug=True)
        while not client.connect():
            print('Ensure end host is running and target ip:port are correct.')
            time.sleep(0.5)
            pass

        if args['files']:
            file_paths = args['<file_path>']
            file_path = file_paths[client.files_sent_len]
            while True:
                if not os.path.exists(file_path):
                    print("Path to file is incorrect: ", file_path)
                    file_paths.pop(client.files_sent_len)

                if client.files_sent_len == len(file_paths):
                    break
                file_path = file_paths[client.files_sent_len]

                try:
                    client.send_file(file_path)
                except KeyboardInterrupt:
                    print('KeyboardInterrupt')
                    break
                except socket.error:  # Any error will trigger a reconnect, be aware for debugging
                    while not client.connect():
                        pass

        elif args['dir']:
            dir_path = args['<dir_path>']
            if dir_path[-1] != '/':
                dir_path += '/'
            if args['--limit'] is not None:
                file_limit = int(args['--limit'])
            else:
                file_limit = 0
            stayalive = args['--stayalive']

            file_history = []
            while True:
                filelist = [''.join([dir_path, f]) for f in os.listdir(dir_path) if os.path.isfile(dir_path + f)]
                new_files = [f for f in filelist if f not in file_history]
                file_history.extend(new_files)

                if new_files:
                    len_newfiles_sent = 0
                    file_path = new_files[len_newfiles_sent]
                    while True:
                        if not os.path.exists(file_path):
                            print("Path to file is incorrect: ", file_path)
                            new_files.pop(client.files_sent_len)

                        if len_newfiles_sent == len(new_files):
                            break
                        elif client.files_sent_len >= file_limit and file_limit:
                            break
                        file_path = new_files[len_newfiles_sent]

                        try:
                            client.send_file(file_path)
                            len_newfiles_sent += 1
                        except KeyboardInterrupt:
                            print('KeyboardInterrupt')
                            break
                        except socket.error:
                            while not client.connect():
                                pass
                if client.files_sent_len >= file_limit and file_limit:
                    break
                elif not stayalive:
                    break

        client.close()
    print("fin.")


if __name__ == '__main__':
    from docopt import docopt
    args = docopt(__doc__)
    main(args)



from __future__ import print_function
import os
import sys
import filecmp
import time
import threading
import shutil

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
import folan

arg_dict = {'listen': False, '<ip:port>': '127.0.0.1:50000', '--save_path': 'folan_dest/', '--stayalive': False,
            '--limit': None, 'send': False, 'files': False, '<file_path>': [], 'dir': False, '<dir_path>': None,
            '--help': False}

def serv(args):
    args['listen'] = True
    folan.main(args)


def cli(args):
    args['send'] = True
    folan.main(args)


def write_dummy_file(filename='temp.txt', filesize=1048576):
    """ Creates 1MB temporary file """
    with open(filename, 'wb') as f:
        size = filesize
        f.write(b"\0" * size)


def test_empty_file_send():
    write_dummy_file(filesize=0)

    serv_args = arg_dict.copy()
    serv_args['listen'] = True
    serv_args['--limit'] = 1
    cli_args = arg_dict.copy()
    cli_args['send'] = True
    cli_args['files'] = True
    cli_args['<file_path>'] = ['temp.txt']

    server_thread = threading.Thread(target=serv, args=(serv_args,))
    client_thread = threading.Thread(target=cli, args=(cli_args,))
    server_thread.start()
    client_thread.start()
    while server_thread.is_alive():
        pass

    assert filecmp.cmp('temp.txt', 'folan_dest/temp.txt', shallow=False)  # file saved in local directory
    os.remove('temp.txt')
    shutil.rmtree('folan_dest/')



from __future__ import print_function
import os
import sys
import filecmp
import time
import threading
import shutil

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
import folan

arg_dict = {'listen': False, '<ip:port>': '127.0.0.1:50000', '--save_path': 'folan_dest/', '--stayalive': False,
            '--limit': None, 'send': False, 'files': False, '<file_path>': [], 'dir': False, '<dir_path>': None,
            '--help': False}

def serv(args):
    args['listen'] = True
    folan.main(args)


def cli(args):
    args['send'] = True
    folan.main(args)


def write_dummy_file(filename='temp.txt', filesize=1048576):
    """ Creates 1MB temporary file """
    with open(filename, 'wb') as f:
        size = filesize
        f.write(b"\0" * size)


def test_local_dir_save():
    if os.path.exists('test_local_dir_save/'):
        shutil.rmtree('test_local_dir_save/')
    os.makedirs('test_local_dir_save/')
    write_dummy_file('test_local_dir_save/temp.txt')
    if os.path.exists('temp.txt'):
        os.remove('temp.txt')

    serv_args = arg_dict.copy()
    serv_args['--save_path'] = '.'
    serv_args['--limit'] = 1
    cli_args = arg_dict.copy()
    cli_args['files'] = True
    cli_args['<file_path>'] = ['test_local_dir_save/temp.txt']

    server_thread = threading.Thread(target=serv, args=(serv_args,))
    client_thread = threading.Thread(target=cli, args=(cli_args,))
    server_thread.start()
    client_thread.start()
    while client_thread.is_alive():
        pass

    assert filecmp.cmp('temp.txt', 'test_local_dir_save/temp.txt', shallow=False)  # file saved in local directory
    os.remove('temp.txt')
    shutil.rmtree('test_local_dir_save/')


def test_directory_send():
    if os.path.exists('test_directory_send/'):
        shutil.rmtree('test_directory_send/')
    os.makedirs('test_directory_send')
    os.makedirs('test_directory_send/send/')
    os.makedirs('test_directory_send/recv/')
    for i in range(3):
        write_dummy_file('test_directory_send/send/{}.txt'.format(str(i)))

    serv_args = arg_dict.copy()
    serv_args['<ip:port>'] = '127.0.0.1:50001'
    serv_args['--save_path'] = 'test_directory_send/recv/'
    serv_args['--limit'] = 10
    cli_args = arg_dict.copy()
    cli_args['<ip:port>'] = '127.0.0.1:50001'
    cli_args['dir'] = True
    cli_args['<dir_path>'] = 'test_directory_send/send/'

    server_thread = threading.Thread(target=serv, args=(serv_args,))
    client_thread = threading.Thread(target=cli, args=(cli_args,))
    server_thread.start()
    client_thread.start()
    while client_thread.is_alive():
        pass

    for i in range(3):
        assert filecmp.cmp('test_directory_send/send/0.txt', 'test_directory_send/recv/{}.txt'.format(str(i)),
                           shallow=False)  # all files sent
    shutil.rmtree('test_directory_send/')


def test_stayalive():
    if os.path.exists('test_stayalive/'):
        shutil.rmtree('test_stayalive/')
    os.makedirs('test_stayalive/send/')
    os.makedirs('test_stayalive/recv/')
    write_dummy_file('test_stayalive/send/before.txt')

    serv_args = arg_dict.copy()
    serv_args['<ip:port>'] = '127.0.0.1:50002'
    serv_args['--save_path'] = 'test_stayalive/recv/'
    serv_args['--limit'] = 2
    cli_args = arg_dict.copy()
    cli_args['<ip:port>'] = '127.0.0.1:50002'
    cli_args['dir'] = True
    cli_args['<dir_path>'] = 'test_stayalive/send/'
    cli_args['--stayalive'] = True
    cli_args['--limit'] = 2

    server_thread = threading.Thread(target=serv, args=(serv_args,))
    client_thread = threading.Thread(target=cli, args=(cli_args,))
    server_thread.start()
    client_thread.start()
    time.sleep(2)  # enough time for one file to send
    write_dummy_file('test_stayalive/send/after.txt')
    while server_thread.is_alive():
        pass

    assert filecmp.cmp('test_stayalive/send/after.txt', 'test_stayalive/recv/after.txt',
                       shallow=False)  # sender stay alive for new file to be sent
    shutil.rmtree('test_stayalive/')


if __name__ == '__main__':
    print("\n\ntest_local_dir_save()")
    test_local_dir_save()
    print("\n\ntest_directory_send()")
    test_directory_send()
    print("\n\ntest_stayalive()")
    test_stayalive()
    print('\n\nfin.')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from folan import __version__
from setuptools import setup
import os
import sys


if sys.argv[-1] == "publish":
    if os.path.exists('dist/'):
        if os.listdir('dist'):
            raise Exception("Clean dist/ first.")
    os.system("python setup.py sdist bdist_wheel --universal")
    os.system("twine upload dist/*")
    sys.exit()

REQUIRED = [
    'docopt==0.6.2'
]

setup(
    name='folan',
    version=__version__,
    author='Matthew Timms',
    author_email='matthewtimms@live.com.au',
    description='Sharing files over LAN',
    license='MIT',
    url='https://github.com/MattTimms/folan',
    py_modules=['folan'],
    install_requires=REQUIRED,
    classifiers=[  # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Communications :: File Sharing',
        'Topic :: Software Development',
        'Topic :: Utilities',
    ],
    entry_points={
        'console_scripts': [
            'folan = folan:main'
        ]
    },
    keywords=['file sharing', 'tcp', 'lan'],\
)
